from __future__ import annotations

import json
from pathlib import Path

import pytest

from axisdb.engine.storage import (
    StoragePaths,
    default_payload,
    read_validated,
    recover_if_needed,
)
from axisdb.errors import StorageCorruptionError


def test_recovery_promotes_tmp_when_main_missing(tmp_path: Path) -> None:
    db_path = tmp_path / "db.json"
    paths = StoragePaths(db_path=db_path)

    tmp_payload = default_payload(dimensions=2)
    paths.tmp_path.write_text(json.dumps(tmp_payload), encoding="utf-8")

    recover_if_needed(paths)

    assert db_path.exists()
    assert not paths.tmp_path.exists()
    assert read_validated(db_path)["meta"]["dimensions"] == 2


def test_recovery_deletes_tmp_when_main_valid(tmp_path: Path) -> None:
    db_path = tmp_path / "db.json"
    paths = StoragePaths(db_path=db_path)

    db_path.write_text(json.dumps(default_payload(dimensions=1)), encoding="utf-8")
    paths.tmp_path.write_text(
        json.dumps(default_payload(dimensions=2)), encoding="utf-8"
    )

    recover_if_needed(paths)

    assert db_path.exists()
    assert not paths.tmp_path.exists()
    assert read_validated(db_path)["meta"]["dimensions"] == 1


def test_recovery_promotes_tmp_when_main_invalid(tmp_path: Path) -> None:
    db_path = tmp_path / "db.json"
    paths = StoragePaths(db_path=db_path)

    db_path.write_text("{not json", encoding="utf-8")
    paths.tmp_path.write_text(
        json.dumps(default_payload(dimensions=3)), encoding="utf-8"
    )

    recover_if_needed(paths)
    assert read_validated(db_path)["meta"]["dimensions"] == 3


def test_recovery_raises_when_no_valid_file_exists(tmp_path: Path) -> None:
    db_path = tmp_path / "db.json"
    paths = StoragePaths(db_path=db_path)

    db_path.write_text("{not json", encoding="utf-8")
    paths.tmp_path.write_text("{also not json", encoding="utf-8")

    with pytest.raises(StorageCorruptionError):
        recover_if_needed(paths)



"""Query evaluation for AxisDB."""

from __future__ import annotations

from collections.abc import Iterable
from typing import Any

from axisdb.engine.keycodec import decode_key
from axisdb.query.ast import And, Expr, Field, KeyDim, Not, Op, Or


def _contains(container: Any, item: Any) -> bool:
    if isinstance(container, str):
        return str(item) in container
    if isinstance(container, dict):
        return item in container
    if isinstance(container, Iterable):
        try:
            return item in container
        except TypeError:
            return False
    return False


def _apply_op(left: Any, op: Op, right: Any) -> bool:
    if op == "==":
        return left == right
    if op == "!=":
        return left != right
    if op == "<":
        return left < right
    if op == "<=":
        return left <= right
    if op == ">":
        return left > right
    if op == ">=":
        return left >= right
    if op == "in":
        return left in right
    if op == "contains":
        return _contains(left, right)
    raise ValueError(f"Unsupported operator: {op}")


def _get_field(value: Any, path: tuple[str, ...]) -> Any:
    cur = value
    for p in path:
        if not isinstance(cur, dict):
            return None
        if p not in cur:
            return None
        cur = cur[p]
    return cur


def evaluate(expr: Expr, *, encoded_key: str, value: Any) -> bool:
    if isinstance(expr, And):
        return evaluate(expr.left, encoded_key=encoded_key, value=value) and evaluate(
            expr.right, encoded_key=encoded_key, value=value
        )
    if isinstance(expr, Or):
        return evaluate(expr.left, encoded_key=encoded_key, value=value) or evaluate(
            expr.right, encoded_key=encoded_key, value=value
        )
    if isinstance(expr, Not):
        return not evaluate(expr.expr, encoded_key=encoded_key, value=value)
    if isinstance(expr, KeyDim):
        key_components = decode_key(encoded_key)
        left = key_components[expr.index]
        return _apply_op(left, expr.op, expr.value)
    if isinstance(expr, Field):
        left = _get_field(value, expr.path)
        return _apply_op(left, expr.op, expr.value)
    raise TypeError(f"Unsupported expression type: {type(expr)}")



"""Public API for AxisDB.

This module intentionally keeps the public surface small and typed.
Engine details live in [`axisdb.engine.storage`](axisdb/engine/storage.py:1),
[`axisdb.engine.locking`](axisdb/engine/locking.py:1), and other `axisdb.*` modules.
"""

from __future__ import annotations

import json
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Literal

from axisdb.engine.keycodec import decode_key, encode_key
from axisdb.engine.locking import FileLock, FileLockSpec, LockMode, LockPaths
from axisdb.engine.storage import (
    FieldIndexDef,
    StoragePaths,
    default_payload,
    read_validated,
    recover_if_needed,
    write_atomic,
)
from axisdb.errors import (
    InvalidCoordsError,
    NonJsonSerializableValueError,
    ReadOnlyError,
    StorageCorruptionError,
    ValidationError,
    WrongDimensionLengthError,
)
from axisdb.index.fields import canonical_value_key, rebuild_field_indexes
from axisdb.index.prefix import rebuild_prefix_keys, select_prefix_range
from axisdb.query.ast import Expr, Field, is_simple_field_equality
from axisdb.query.eval import evaluate

OpenMode = Literal["r", "rw"]


@dataclass
class AxisDB:
    """Main database handle.

    Implementation is added incrementally; this file is the stable public
    surface that other interfaces (CLI, server) will use.
    """

    path: Path
    mode: OpenMode = "rw"
    lock: bool = True

    _dimensions: int = field(init=False)
    _base_data: dict[str, Any] = field(init=False, default_factory=dict)
    _base_prefix_keys: list[str] = field(init=False, default_factory=list)
    _base_field_indexes: dict[str, dict[str, list[str]]] = field(
        init=False, default_factory=dict
    )
    _field_index_defs: list[FieldIndexDef] = field(init=False, default_factory=list)

    _overlay_set: dict[str, Any] = field(init=False, default_factory=dict)
    _overlay_del: set[str] = field(init=False, default_factory=set)

    _writer_lock: FileLock | None = field(init=False, default=None)
    _lock_paths: LockPaths = field(init=False)
    _storage_paths: StoragePaths = field(init=False)

    @classmethod
    def open(cls, path: str | Path, mode: OpenMode = "rw", lock: bool = True) -> AxisDB:
        db = cls(path=Path(path), mode=mode, lock=lock)
        db._initialize()
        return db

    @classmethod
    def create(
        cls,
        path: str | Path,
        *,
        dimensions: int,
        overwrite: bool = False,
        lock: bool = True,
    ) -> AxisDB:
        """Create a new database file.

        This is a convenience helper (not part of the minimal API shape in the spec),
        required to bootstrap a database because dimensions are fixed at creation time.
        """

        p = Path(path)
        if p.exists() and not overwrite:
            raise ValidationError(f"Database already exists: {p}")

        storage_paths = StoragePaths(db_path=p)
        payload = default_payload(dimensions=dimensions)
        write_atomic(storage_paths, payload)
        return cls.open(p, mode="rw", lock=lock)

    @property
    def dimensions(self) -> int:
        return self._dimensions

    def define_field_index(self, name: str, path: tuple[str, ...] | list[str]) -> None:
        """Define a field index stored in DB metadata and rebuilt on commit.

        This is a correctness-first MVP feature:
        - Indexes are rebuilt on each commit.
        - `find()` can use the index for simple `Field(path) == literal` queries.
        """

        self._assert_writable()
        if not name or not isinstance(name, str):
            raise ValidationError("index name must be a non-empty string")
        if not isinstance(path, (list, tuple)) or not all(
            isinstance(p, str) for p in path
        ):
            raise ValidationError("index path must be a list/tuple of strings")

        idx: FieldIndexDef = {"name": name, "path": list(path)}

        # Replace existing definition with same name.
        self._field_index_defs = [
            d for d in self._field_index_defs if d["name"] != name
        ]
        self._field_index_defs.append(idx)

    def __enter__(self) -> AxisDB:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: object | None,
    ) -> None:
        if self.mode == "rw":
            # Safety: never auto-commit.
            self.rollback()
        if self._writer_lock is not None:
            self._writer_lock.__exit__(exc_type, exc, tb)
            self._writer_lock = None
        return None

    # ---------------------------------------------------------------------
    # Initialization
    # ---------------------------------------------------------------------

    def _initialize(self) -> None:
        self._lock_paths = LockPaths(self.path)
        self._storage_paths = StoragePaths(db_path=self.path)

        # If locking is enabled and we're a writer, enforce single-writer.
        if self.lock and self.mode == "rw":
            self._writer_lock = FileLock(
                # Fail fast if another writer already exists.
                FileLockSpec(
                    self._lock_paths.writer_lock,
                    LockMode.EXCLUSIVE,
                    timeout_s=0.0,
                )
            )
            self._writer_lock.__enter__()

        # Recovery should be performed with exclusive rw lock to avoid races.
        if self.lock:
            with FileLock(FileLockSpec(self._lock_paths.rw_lock, LockMode.EXCLUSIVE)):
                recover_if_needed(self._storage_paths)
        else:
            recover_if_needed(self._storage_paths)

        self._reload_base_from_disk()

    def _reload_base_from_disk(self) -> None:
        if not self.path.exists():
            raise StorageCorruptionError(f"Database file does not exist: {self.path}")

        if self.lock:
            lock_mode = LockMode.SHARED
            with FileLock(FileLockSpec(self._lock_paths.rw_lock, lock_mode)):
                payload = read_validated(self.path)
        else:
            payload = read_validated(self.path)

        self._dimensions = int(payload["meta"]["dimensions"])
        self._base_data = dict(payload.get("data", {}))
        index_payload = payload.get("index", {})
        self._base_prefix_keys = list(index_payload.get("prefix_keys", []))
        self._base_field_indexes = dict(index_payload.get("fields", {}))
        self._field_index_defs = list(
            payload["meta"].get("indexes", {}).get("fields", [])
        )

        # Clear overlays after a reload.
        self._overlay_set.clear()
        self._overlay_del.clear()

    def _assert_writable(self) -> None:
        if self.mode != "rw":
            raise ReadOnlyError("Database opened in read-only mode")

    def _assert_key(self, key: tuple[str, ...]) -> None:
        if not isinstance(key, tuple):
            raise InvalidCoordsError("Key must be a tuple[str, ...]")
        if not all(isinstance(c, str) for c in key):
            raise InvalidCoordsError("All key components must be strings")
        if len(key) != self._dimensions:
            raise WrongDimensionLengthError(
                f"Expected {self._dimensions} key components, got {len(key)}"
            )

    # ---------------------------------------------------------------------
    # CRUD
    # ---------------------------------------------------------------------

    def get(self, key: tuple[str, ...]) -> Any:
        self._assert_key(key)
        ek = encode_key(key)

        if ek in self._overlay_del:
            raise KeyError(key)
        if ek in self._overlay_set:
            return self._overlay_set[ek]

        if self.mode == "r":
            # Read-only uses fresh reads per call.
            self._reload_base_from_disk()

        return self._base_data[ek]

    def set(self, key: tuple[str, ...], value: Any) -> None:
        self._assert_writable()
        self._assert_key(key)

        # Ensure JSON-serializable by default.
        try:
            json.dumps(value)
        except TypeError as exc:
            raise NonJsonSerializableValueError(
                "Value is not JSON-serializable"
            ) from exc

        ek = encode_key(key)
        self._overlay_set[ek] = value
        self._overlay_del.discard(ek)

    def delete(self, key: tuple[str, ...]) -> None:
        self._assert_writable()
        self._assert_key(key)
        ek = encode_key(key)

        if ek in self._overlay_set:
            del self._overlay_set[ek]
        self._overlay_del.add(ek)

    def exists(self, key: tuple[str, ...]) -> bool:
        self._assert_key(key)
        ek = encode_key(key)
        if ek in self._overlay_del:
            return False
        if ek in self._overlay_set:
            return True
        if self.mode == "r":
            self._reload_base_from_disk()
        return ek in self._base_data

    # ---------------------------------------------------------------------
    # Listing / Querying
    # ---------------------------------------------------------------------

    def list(
        self, prefix: tuple[str, ...] | None = None, depth: int | None = None
    ) -> list[tuple[str, ...]]:
        """List keys.

        - `prefix` filters by key-prefix.
        - `depth` limits returned key length relative to prefix.
        """

        if prefix is None:
            prefix = ()
        if len(prefix) > self._dimensions:
            raise ValidationError("prefix longer than number of dimensions")

        if self.mode == "r":
            self._reload_base_from_disk()

        keys = self._materialized_keys()
        if prefix:
            ep = encode_key(prefix)
            keys = [k for k in keys if k.startswith(ep + "/") or k == ep]

        out: list[tuple[str, ...]] = []
        for ek in keys:
            dk = decode_key(ek)
            if depth is not None:
                max_len = len(prefix) + depth
                dk = dk[:max_len]
            out.append(dk)

        # Deduplicate if depth truncation caused collisions.
        return sorted(set(out))

    def slice(self, dim_slices: object) -> object:
        """Return a nested dict slice based on per-dimension selectors.

        `dim_slices` is a sequence with up to N elements (N == dimensions).
        If shorter than N, it is padded with `None` (wildcards).

        Each selector may be:
        - None: wildcard
        - str: exact match
        - list/tuple/set of str: membership match
        - callable: predicate on the dimension component (slower)
        """

        if self.mode == "r":
            self._reload_base_from_disk()

        selectors = self._normalize_dim_slices(dim_slices)

        out: dict[str, Any] = {}
        for encoded_key in self._materialized_keys():
            key = decode_key(encoded_key)
            if not self._match_dim_slices(selectors, key):
                continue

            node: dict[str, Any] = out
            for comp in key[:-1]:
                nxt = node.get(comp)
                if not isinstance(nxt, dict):
                    nxt = {}
                    node[comp] = nxt
                node = nxt
            node[key[-1]] = self._get_by_encoded_key(encoded_key)

        return out

    def _normalize_dim_slices(self, dim_slices: object) -> tuple[object, ...]:
        if not isinstance(dim_slices, (list, tuple)):
            raise ValidationError("dim_slices must be a list/tuple")
        if len(dim_slices) > self._dimensions:
            raise ValidationError("dim_slices longer than number of dimensions")
        padded = tuple(dim_slices) + (None,) * (self._dimensions - len(dim_slices))
        return padded

    def _match_dim_slices(
        self, selectors: tuple[object, ...], key: tuple[str, ...]
    ) -> bool:
        for selector, comp in zip(selectors, key, strict=True):
            if selector is None:
                continue
            if isinstance(selector, str):
                if comp != selector:
                    return False
                continue
            if isinstance(selector, (list, tuple, set)):
                if comp not in selector:
                    return False
                continue
            if callable(selector):
                if not bool(selector(comp)):
                    return False
                continue
            raise ValidationError(
                "Invalid dim_slices selector (expected None, str, list/tuple/set, or callable)"
            )
        return True

    def find(
        self,
        prefix: tuple[str, ...] | None = None,
        where: Expr | None | object = None,
        limit: int | None = None,
    ) -> list[tuple[tuple[str, ...], Any]]:
        if prefix is None:
            prefix = ()
        if len(prefix) > self._dimensions:
            raise ValidationError("prefix longer than number of dimensions")
        if limit is not None and limit <= 0:
            raise ValidationError("limit must be a positive integer")

        if self.mode == "r":
            self._reload_base_from_disk()

        candidates = self._candidate_keys(prefix=prefix, where=where)
        results: list[tuple[tuple[str, ...], Any]] = []
        for ek in candidates:
            doc = self._get_by_encoded_key(ek)
            if where is None:
                ok = True
            elif isinstance(where, Expr):
                ok = evaluate(where, encoded_key=ek, value=doc)
            elif callable(where):
                ok = bool(where(doc))
            else:
                raise ValidationError("where must be an Expr or callable")

            if ok:
                results.append((decode_key(ek), doc))
                if limit is not None and len(results) >= limit:
                    break
        return results

    def _candidate_keys(self, prefix: tuple[str, ...], where: object) -> list[str]:
        """Return candidate encoded keys.

        Optimizations (correctness-first):
        - Uses prefix index when available.
        - Uses field index for simple `Field(path) == literal` queries when configured.
        """

        all_keys = self._materialized_keys()

        # Base-only index usage is only safe when there is no overlay.
        can_use_indexes = (
            not self._overlay_set and not self._overlay_del and self._base_prefix_keys
        )

        # 1) Field index (if enabled and query is simple equality)
        if (
            can_use_indexes
            and isinstance(where, Expr)
            and is_simple_field_equality(where)
            and isinstance(where, Field)
        ):
            # Find the configured index with an identical path.
            index_name = None
            for d in self._field_index_defs:
                if tuple(d["path"]) == tuple(where.path):
                    index_name = d["name"]
                    break

            if index_name is not None:
                vkey = canonical_value_key(where.value)
                candidates = self._base_field_indexes.get(index_name, {}).get(vkey, [])

                # Apply prefix restriction on top.
                if prefix:
                    ep = encode_key(prefix)
                    return [k for k in candidates if k.startswith(ep + "/") or k == ep]
                return list(candidates)

        # 2) Prefix index
        if prefix:
            if can_use_indexes:
                ep = encode_key(prefix)
                lo, hi = select_prefix_range(self._base_prefix_keys, ep)
                return self._base_prefix_keys[lo:hi]

            ep = encode_key(prefix)
            return [k for k in all_keys if k.startswith(ep + "/") or k == ep]

        return all_keys

    def _materialized_keys(self) -> list[str]:
        keys = set(self._base_data.keys())
        keys.difference_update(self._overlay_del)
        keys.update(self._overlay_set.keys())
        return sorted(keys)

    def _get_by_encoded_key(self, encoded_key: str) -> Any:
        if encoded_key in self._overlay_del:
            raise KeyError(encoded_key)
        if encoded_key in self._overlay_set:
            return self._overlay_set[encoded_key]
        return self._base_data[encoded_key]

    # ---------------------------------------------------------------------
    # Transactions
    # ---------------------------------------------------------------------

    def commit(self) -> None:
        self._assert_writable()

        # Nothing to do.
        if not self._overlay_set and not self._overlay_del:
            return

        # Build next state.
        next_data = dict(self._base_data)
        for ek in self._overlay_del:
            next_data.pop(ek, None)
        next_data.update(self._overlay_set)

        # Rebuild indexes (correctness-first).
        prefix_keys = rebuild_prefix_keys(next_data)
        field_indexes = rebuild_field_indexes(next_data, self._field_index_defs)

        payload = default_payload(dimensions=self._dimensions)
        payload["meta"]["indexes"]["fields"] = self._field_index_defs
        payload["data"] = next_data
        payload["index"] = {"prefix_keys": prefix_keys, "fields": field_indexes}

        if self.lock:
            with FileLock(FileLockSpec(self._lock_paths.rw_lock, LockMode.EXCLUSIVE)):
                write_atomic(self._storage_paths, payload)
        else:
            write_atomic(self._storage_paths, payload)

        # Refresh base and clear overlay.
        self._base_data = next_data
        self._base_prefix_keys = prefix_keys
        self._base_field_indexes = field_indexes
        self._overlay_set.clear()
        self._overlay_del.clear()

    def rollback(self) -> None:
        if self.mode == "r":
            return
        self._reload_base_from_disk()



"""On-disk storage for AxisDB.

Design goals:

- Human-inspectable JSON on disk.
- Atomic commits via temp-file + `os.replace()`.
- Crash recovery on open by promoting a valid temp file.
"""

from __future__ import annotations

import json
import os
from contextlib import suppress
from dataclasses import dataclass
from datetime import UTC, datetime
from pathlib import Path
from typing import Any, TypedDict, cast

from axisdb.errors import StorageCorruptionError, ValidationError

FORMAT_NAME = "axisdb"
FORMAT_VERSION = 2


def _utc_now_iso() -> str:
    return (
        datetime.now(tz=UTC).replace(microsecond=0).isoformat().replace("+00:00", "Z")
    )


class FieldIndexDef(TypedDict):
    name: str
    path: list[str]


class IndexMeta(TypedDict, total=False):
    prefix: dict[str, Any]
    fields: list[FieldIndexDef]


class Meta(TypedDict):
    dimensions: int
    created_at: str
    updated_at: str
    indexes: IndexMeta


class IndexPayload(TypedDict, total=False):
    prefix_keys: list[str]
    fields: dict[str, dict[str, list[str]]]


class DBPayload(TypedDict):
    format: str
    format_version: int
    meta: Meta
    data: dict[str, Any]
    index: IndexPayload


@dataclass(frozen=True)
class StoragePaths:
    db_path: Path

    @property
    def tmp_path(self) -> Path:
        return self.db_path.with_suffix(self.db_path.suffix + ".tmp")


def default_payload(dimensions: int) -> DBPayload:
    if dimensions <= 0:
        raise ValidationError("dimensions must be a positive integer")

    now = _utc_now_iso()
    return {
        "format": FORMAT_NAME,
        "format_version": FORMAT_VERSION,
        "meta": {
            "dimensions": dimensions,
            "created_at": now,
            "updated_at": now,
            "indexes": {"prefix": {"enabled": True}, "fields": []},
        },
        "data": {},
        "index": {"prefix_keys": [], "fields": {}},
    }


def _is_dict(value: Any) -> bool:
    return isinstance(value, dict)


def validate_payload(raw: Any) -> DBPayload:
    if not _is_dict(raw):
        raise ValidationError("DB file must contain a JSON object")

    fmt = raw.get("format")
    ver = raw.get("format_version")
    if fmt != FORMAT_NAME or ver != FORMAT_VERSION:
        raise ValidationError("Unsupported DB format or version")

    meta = raw.get("meta")
    if not _is_dict(meta):
        raise ValidationError("Missing or invalid meta")

    try:
        dimensions = int(meta["dimensions"])
    except Exception as exc:  # noqa: BLE001
        raise ValidationError("Missing or invalid meta.dimensions") from exc
    if dimensions <= 0:
        raise ValidationError("meta.dimensions must be a positive integer")

    for k in ("created_at", "updated_at"):
        if not isinstance(meta.get(k), str):
            raise ValidationError(f"Missing or invalid meta.{k}")

    indexes = meta.get("indexes")
    if indexes is None:
        indexes = {"prefix": {"enabled": True}, "fields": []}
        meta["indexes"] = indexes
    if not _is_dict(indexes):
        raise ValidationError("Missing or invalid meta.indexes")

    data = raw.get("data")
    if not _is_dict(data):
        raise ValidationError("Missing or invalid data")

    index = raw.get("index")
    if index is None:
        index = {"prefix_keys": [], "fields": {}}
        raw["index"] = index
    if not _is_dict(index):
        raise ValidationError("Missing or invalid index")

    # Soft-validate index structure for MVP: allow empty/missing fields.
    if "prefix_keys" in index and not isinstance(index["prefix_keys"], list):
        raise ValidationError("index.prefix_keys must be a list")
    if "fields" in index and not _is_dict(index["fields"]):
        raise ValidationError("index.fields must be an object")

    # Cast after validation.
    return cast(DBPayload, raw)


def _read_json(path: Path) -> Any:
    try:
        text = path.read_text(encoding="utf-8")
    except FileNotFoundError:
        raise
    except OSError as exc:
        raise StorageCorruptionError(f"Could not read DB file: {path}") from exc
    try:
        return json.loads(text)
    except json.JSONDecodeError as exc:
        raise StorageCorruptionError(f"Invalid JSON in DB file: {path}") from exc


def read_validated(path: Path) -> DBPayload:
    return validate_payload(_read_json(path))


def recover_if_needed(paths: StoragePaths) -> None:
    """Best-effort recovery.

    Rules:
    - If main is missing and tmp exists+valid: promote tmp.
    - If main invalid and tmp exists+valid: promote tmp.
    - If main valid and tmp exists: delete tmp.
    """

    main_exists = paths.db_path.exists()
    tmp_exists = paths.tmp_path.exists()

    if not main_exists and not tmp_exists:
        return

    main_ok = False
    if main_exists:
        try:
            read_validated(paths.db_path)
            main_ok = True
        except (ValidationError, StorageCorruptionError):
            main_ok = False

    tmp_ok = False
    if tmp_exists:
        try:
            read_validated(paths.tmp_path)
            tmp_ok = True
        except (ValidationError, StorageCorruptionError):
            tmp_ok = False

    if main_ok:
        if tmp_exists:
            with suppress(OSError):
                paths.tmp_path.unlink()
        return

    if tmp_ok:
        os.replace(paths.tmp_path, paths.db_path)
        return

    if main_exists:
        raise StorageCorruptionError(
            f"DB file is corrupted and no valid recovery file exists: {paths.db_path}"
        )
    raise StorageCorruptionError(
        f"DB file is missing and temp file is not a valid recovery: {paths.db_path}"
    )


def _fsync_file(f) -> None:  # noqa: ANN001
    f.flush()
    os.fsync(f.fileno())


def _fsync_dir_best_effort(path: Path) -> None:
    """Attempt to fsync the directory containing `path`.

    On Windows this may not be supported; treat it as best-effort.
    """

    directory = path.parent
    try:
        fd = os.open(str(directory), os.O_RDONLY)
    except OSError:
        return
    try:
        os.fsync(fd)
    except OSError:
        return
    finally:
        os.close(fd)


def write_atomic(paths: StoragePaths, payload: DBPayload) -> None:
    """Write payload to tmp and atomically replace main."""

    payload["meta"]["updated_at"] = _utc_now_iso()

    tmp = paths.tmp_path
    tmp.parent.mkdir(parents=True, exist_ok=True)

    encoded = json.dumps(payload, indent=2, sort_keys=True, ensure_ascii=False) + "\n"
    try:
        with open(tmp, "w", encoding="utf-8", newline="\n") as f:
            f.write(encoded)
            _fsync_file(f)
    except OSError as exc:
        raise StorageCorruptionError(f"Failed to write temp DB file: {tmp}") from exc

    os.replace(tmp, paths.db_path)
    _fsync_dir_best_effort(paths.db_path)



from __future__ import annotations

from pathlib import Path

from axisdb import AxisDB
from axisdb.query.ast import Field


def test_find_uses_field_index_for_simple_equality_with_prefix(tmp_path: Path) -> None:
    db_path = tmp_path / "db.json"
    db = AxisDB.create(db_path, dimensions=2)

    # Configure a field index on customer_id.
    db.define_field_index("by_customer_id", ("customer_id",))

    db.set(("orders", "1"), {"customer_id": "c1", "amount": 10})
    db.set(("orders", "2"), {"customer_id": "c2", "amount": 20})
    db.set(("events", "1"), {"customer_id": "c2", "kind": "click"})
    db.commit()

    expr = Field(("customer_id",), "==", "c2")
    rows = db.find(prefix=("orders",), where=expr)
    assert rows == [(("orders", "2"), {"customer_id": "c2", "amount": 20})]


def test_find_falls_back_when_no_matching_index(tmp_path: Path) -> None:
    db_path = tmp_path / "db.json"
    db = AxisDB.create(db_path, dimensions=1)

    db.set(("1",), {"customer_id": "c1"})
    db.set(("2",), {"customer_id": "c2"})
    db.commit()

    expr = Field(("customer_id",), "==", "c2")
    rows = db.find(where=expr)
    assert rows == [(("2",), {"customer_id": "c2"})]



from __future__ import annotations

import pytest

from axisdb.engine.keycodec import decode_key, encode_key
from axisdb.errors import ValidationError


@pytest.mark.parametrize(
    "components",
    [
        ("a",),
        ("a", "b"),
        ("user1", "2025-01", "orders"),
        ("with space", "with/slash", "with%percent"),
        ("", "empty", "component"),
    ],
)
def test_encode_decode_roundtrip(components: tuple[str, ...]) -> None:
    assert decode_key(encode_key(components)) == components


def test_decode_empty_string_is_empty_tuple() -> None:
    assert decode_key("") == ()


def test_encode_rejects_non_string_component() -> None:
    with pytest.raises(ValidationError):
        encode_key(("ok", 123))  # type: ignore[arg-type]



"""Cross-platform file locking primitives.

The storage design requires:

- A single-writer lock held for the lifetime of a writer session.
- A read/write lock used as shared during reads and exclusive during commits.
"""

from __future__ import annotations

import os
import time
from contextlib import AbstractContextManager, suppress
from dataclasses import dataclass
from enum import Enum
from pathlib import Path

import portalocker

from axisdb.errors import LockError


class LockMode(str, Enum):
    SHARED = "shared"
    EXCLUSIVE = "exclusive"


@dataclass(frozen=True)
class FileLockSpec:
    path: Path
    mode: LockMode
    timeout_s: float | None = 10.0


class FileLock(AbstractContextManager["FileLock"]):
    """A context-managed file lock using portalocker."""

    def __init__(self, spec: FileLockSpec) -> None:
        self._spec = spec
        self._fh = None

    def __enter__(self) -> FileLock:
        flags = (
            portalocker.LOCK_SH
            if self._spec.mode == LockMode.SHARED
            else portalocker.LOCK_EX
        )
        flags |= portalocker.LOCK_NB

        self._spec.path.parent.mkdir(parents=True, exist_ok=True)

        # Open our own handle so we can control retry behavior.
        # Avoid append-mode on Windows because some lock implementations are
        # sensitive to how the handle was opened.
        if self._spec.path.exists():
            self._fh = open(self._spec.path, "r+b")
        else:
            self._fh = open(self._spec.path, "w+b")
            # Ensure non-empty content.
            self._fh.write(b"0")
            self._fh.flush()
            with suppress(OSError):
                os.fsync(self._fh.fileno())

        self._fh.seek(0)

        deadline = None
        if self._spec.timeout_s is not None:
            deadline = time.monotonic() + float(self._spec.timeout_s)

        while True:
            try:
                portalocker.lock(self._fh, flags)
                break
            except portalocker.exceptions.LockException as exc:
                if deadline is None or time.monotonic() >= deadline:
                    self._fh.close()
                    self._fh = None
                    raise LockError(
                        f"Could not acquire lock: {self._spec.path}"
                    ) from exc
                time.sleep(0.05)

        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: object | None,
    ) -> None:
        if self._fh is not None:
            try:
                portalocker.unlock(self._fh)
            finally:
                self._fh.close()
                self._fh = None


@dataclass(frozen=True)
class LockPaths:
    """Derive lock file paths from the main database path."""

    db_path: Path

    @property
    def writer_lock(self) -> Path:
        return self.db_path.with_suffix(self.db_path.suffix + ".writer.lock")

    @property
    def rw_lock(self) -> Path:
        return self.db_path.with_suffix(self.db_path.suffix + ".rw.lock")



"""Field/value indexes.

MVP approach: rebuild indexes on each commit for correctness and simplicity.
"""

from __future__ import annotations

import json
from typing import Any

from axisdb.engine.storage import FieldIndexDef


def canonical_value_key(value: Any) -> str:
    return json.dumps(value, sort_keys=True, separators=(",", ":"), ensure_ascii=False)


def _extract_field(value: Any, path: list[str]) -> Any:
    cur = value
    for p in path:
        if not isinstance(cur, dict) or p not in cur:
            return None
        cur = cur[p]
    return cur


def rebuild_field_indexes(
    data: dict[str, Any], index_defs: list[FieldIndexDef]
) -> dict[str, dict[str, list[str]]]:
    out: dict[str, dict[str, list[str]]] = {}
    for d in index_defs:
        out[d["name"]] = {}

    for encoded_key, doc in data.items():
        for d in index_defs:
            extracted = _extract_field(doc, d["path"])
            if extracted is None:
                continue
            vkey = canonical_value_key(extracted)
            bucket = out[d["name"]].setdefault(vkey, [])
            bucket.append(encoded_key)

    # Stable ordering for test determinism.
    for idx in out.values():
        for bucket in idx.values():
            bucket.sort()
    return out



"""Reversible encoding for N-dimensional coordinate keys.

Keys are represented externally as sequences of strings (components).
On disk we store a single encoded string using '/' separators.

We percent-encode each component with no safe characters to ensure
reversibility even when components contain '/', '%', or whitespace.
"""

from __future__ import annotations

from collections.abc import Iterable, Sequence
from urllib.parse import quote, unquote

from axisdb.errors import ValidationError

_SEP = "/"


def encode_component(component: str) -> str:
    if not isinstance(component, str):
        raise ValidationError("Key components must be strings")
    # safe="" ensures '/' is always encoded.
    return quote(component, safe="")


def decode_component(encoded: str) -> str:
    if not isinstance(encoded, str):
        raise ValidationError("Encoded key components must be strings")
    return unquote(encoded)


def encode_key(components: Sequence[str]) -> str:
    if not isinstance(components, Sequence):
        raise ValidationError("Key must be a sequence of strings")
    return _SEP.join(encode_component(c) for c in components)


def decode_key(encoded_key: str) -> tuple[str, ...]:
    if not isinstance(encoded_key, str):
        raise ValidationError("Encoded key must be a string")
    if encoded_key == "":
        return ()
    return tuple(decode_component(part) for part in encoded_key.split(_SEP))


def encode_prefix(prefix: Iterable[str]) -> str:
    """Encode a prefix (0..N components) into an encoded key prefix."""

    return encode_key(tuple(prefix))



"""Query expression tree (AST)."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Literal

Op = Literal["==", "!=", "<", "<=", ">", ">=", "in", "contains"]


@dataclass(frozen=True)
class Expr:
    pass


@dataclass(frozen=True)
class And(Expr):
    left: Expr
    right: Expr


@dataclass(frozen=True)
class Or(Expr):
    left: Expr
    right: Expr


@dataclass(frozen=True)
class Not(Expr):
    expr: Expr


@dataclass(frozen=True)
class KeyDim(Expr):
    index: int
    op: Op
    value: Any


@dataclass(frozen=True)
class Field(Expr):
    path: tuple[str, ...]
    op: Op
    value: Any


def is_simple_field_equality(expr: Expr) -> bool:
    return isinstance(expr, Field) and expr.op == "=="



from __future__ import annotations

from pathlib import Path

import pytest

from axisdb import AxisDB
from axisdb.errors import (
    InvalidCoordsError,
    NonJsonSerializableValueError,
    ReadOnlyError,
    WrongDimensionLengthError,
)
from axisdb.query.ast import Field


def test_create_set_get_commit_roundtrip(tmp_path: Path) -> None:
    db_path = tmp_path / "db.json"
    db = AxisDB.create(db_path, dimensions=2)
    db.set(("u1", "k1"), {"v": 1})
    db.commit()

    db2 = AxisDB.open(db_path, mode="r")
    assert db2.get(("u1", "k1")) == {"v": 1}


def test_rollback_discards_overlay(tmp_path: Path) -> None:
    db_path = tmp_path / "db.json"
    db = AxisDB.create(db_path, dimensions=1)
    db.set(("a",), 1)
    db.commit()

    db.set(("a",), 2)
    db.rollback()
    assert db.get(("a",)) == 1


def test_read_only_rejects_writes(tmp_path: Path) -> None:
    db_path = tmp_path / "db.json"
    AxisDB.create(db_path, dimensions=1)

    ro = AxisDB.open(db_path, mode="r")
    with pytest.raises(ReadOnlyError):
        ro.set(("a",), 1)


def test_set_rejects_non_json_serializable_value(tmp_path: Path) -> None:
    db_path = tmp_path / "db.json"
    db = AxisDB.create(db_path, dimensions=1)
    with pytest.raises(NonJsonSerializableValueError):
        db.set(("a",), {"x": {1, 2, 3}})


def test_rejects_wrong_dimension_length(tmp_path: Path) -> None:
    db_path = tmp_path / "db.json"
    db = AxisDB.create(db_path, dimensions=2)
    with pytest.raises(WrongDimensionLengthError):
        db.get(("only_one",))


def test_rejects_invalid_coords_type(tmp_path: Path) -> None:
    db_path = tmp_path / "db.json"
    db = AxisDB.create(db_path, dimensions=1)
    with pytest.raises(InvalidCoordsError):
        db.get(["a"])  # type: ignore[arg-type]


def test_list_with_prefix_and_depth(tmp_path: Path) -> None:
    db_path = tmp_path / "db.json"
    db = AxisDB.create(db_path, dimensions=3)
    db.set(("u1", "2025", "01"), 1)
    db.set(("u1", "2025", "02"), 2)
    db.set(("u2", "2025", "01"), 3)
    db.commit()

    assert set(db.list(prefix=("u1",))) == {
        ("u1", "2025", "01"),
        ("u1", "2025", "02"),
    }
    # Depth relative to prefix: prefix len=1, depth=1 => return 2 components.
    assert set(db.list(prefix=("u1",), depth=1)) == {("u1", "2025")}


def test_find_with_field_predicate(tmp_path: Path) -> None:
    db_path = tmp_path / "db.json"
    db = AxisDB.create(db_path, dimensions=2)
    db.set(("k", "1"), {"customer_id": "c1", "amount": 10})
    db.set(("k", "2"), {"customer_id": "c2", "amount": 20})
    db.commit()

    expr = Field(("customer_id",), "==", "c2")
    rows = db.find(prefix=("k",), where=expr)
    assert rows == [(("k", "2"), {"customer_id": "c2", "amount": 20})]



"""Prefix index.

This is a correctness-first implementation:

- The on-disk prefix index stores all encoded keys in sorted order.
- Prefix scans become a bisect range query.
"""

from __future__ import annotations

from bisect import bisect_left
from typing import Any


def rebuild_prefix_keys(data: dict[str, Any]) -> list[str]:
    return sorted(data.keys())


def select_prefix_range(prefix_keys: list[str], encoded_prefix: str) -> tuple[int, int]:
    """Return [lo, hi) range of keys starting with encoded_prefix.

    We approximate the end bound by choosing a lexicographic successor string.
    For ASCII-safe encodings this works well.

    Note: this is correct when keys are restricted to a stable encoding that
    does not introduce characters below the sentinel.
    """

    if encoded_prefix == "":
        return 0, len(prefix_keys)

    lo = bisect_left(prefix_keys, encoded_prefix)
    # '' is DEL, lexicographically higher than typical URL-escaped chars.
    hi = bisect_left(prefix_keys, encoded_prefix + "\u007f")
    return lo, hi



"""Pydantic request/response schemas for the FastAPI wrapper."""

from __future__ import annotations

from typing import Any

from pydantic import BaseModel, Field


class InitResponse(BaseModel):
    path: str
    dimensions: int
    created: bool


class ItemBody(BaseModel):
    coords: list[str] = Field(..., description="N-dimensional coordinate key")
    value: Any


class DeleteBody(BaseModel):
    coords: list[str]



from __future__ import annotations

from pathlib import Path

from axisdb import AxisDB


def test_slice_with_exact_match_and_wildcards(tmp_path: Path) -> None:
    db_path = tmp_path / "db.json"
    db = AxisDB.create(db_path, dimensions=3)

    db.set(("u1", "2025", "01"), {"v": 1})
    db.set(("u1", "2025", "02"), {"v": 2})
    db.set(("u2", "2025", "01"), {"v": 3})
    db.commit()

    sliced = db.slice(("u1", None, None))
    assert sliced == {"u1": {"2025": {"01": {"v": 1}, "02": {"v": 2}}}}


def test_slice_with_membership_selector(tmp_path: Path) -> None:
    db_path = tmp_path / "db.json"
    db = AxisDB.create(db_path, dimensions=2)

    db.set(("a", "1"), 1)
    db.set(("a", "2"), 2)
    db.set(("b", "1"), 3)
    db.commit()

    sliced = db.slice(({"a", "b"}, {"1"}))
    assert sliced == {"a": {"1": 1}, "b": {"1": 3}}



"""Public error types for AxisDB."""

from __future__ import annotations


class AxisDBError(Exception):
    """Base class for all AxisDB errors."""


class StorageCorruptionError(AxisDBError):
    """Raised when on-disk data is missing or cannot be validated/recovered."""


class LockError(AxisDBError):
    """Raised when the database cannot acquire the required file locks."""


class ReadOnlyError(AxisDBError):
    """Raised when attempting to mutate state in read-only mode."""


class ValidationError(AxisDBError, ValueError):
    """Raised when inputs fail validation (keys, values, schema)."""


class InvalidCoordsError(ValidationError):
    """Raised when a key/coords value is not a valid coordinate sequence."""


class WrongDimensionLengthError(ValidationError):
    """Raised when a key has the wrong number of dimensions for this database."""


class NonJsonSerializableValueError(ValidationError):
    """Raised when a value cannot be JSON-serialized by the default serializer."""



from __future__ import annotations

import multiprocessing as mp
import time
from pathlib import Path

from axisdb import AxisDB
from axisdb.errors import LockError


def _writer_hold_open(db_path: str, hold_s: float) -> None:
    # Hold the writer lock for a bit.
    db = AxisDB.open(db_path, mode="rw")
    time.sleep(hold_s)
    db.rollback()


def _writer_try_open(db_path: str, q: mp.Queue) -> None:
    try:
        db = AxisDB.open(db_path, mode="rw")
        db.rollback()
        q.put(True)
    except LockError:
        q.put(False)


def test_two_writers_cannot_open_concurrently(tmp_path: Path) -> None:
    db_path = tmp_path / "db.json"
    AxisDB.create(db_path, dimensions=1)

    q: mp.Queue = mp.Queue()

    p1 = mp.Process(target=_writer_hold_open, args=(str(db_path), 1.5))
    p2 = mp.Process(target=_writer_try_open, args=(str(db_path), q))

    p1.start()
    # Give the first writer time to acquire the lock.
    time.sleep(0.2)
    p2.start()

    p2.join(timeout=5)
    p1.join(timeout=5)

    assert q.get(timeout=2) is False


def _reader_get(db_path: str, q: mp.Queue) -> None:
    db = AxisDB.open(db_path, mode="r")
    q.put(db.get(("a",)))


def test_reader_can_open_while_writer_session_exists(tmp_path: Path) -> None:
    db_path = tmp_path / "db.json"
    db = AxisDB.create(db_path, dimensions=1)
    db.set(("a",), 123)
    db.commit()

    writer = mp.Process(target=_writer_hold_open, args=(str(db_path), 1.0))
    q: mp.Queue = mp.Queue()
    reader = mp.Process(target=_reader_get, args=(str(db_path), q))

    writer.start()
    time.sleep(0.2)
    reader.start()

    reader.join(timeout=5)
    writer.join(timeout=5)

    assert q.get(timeout=2) == 123



"""FastAPI wrapper over the AxisDB library.

This is intentionally a thin layer:
- It does not implement database logic.
- It converts HTTP requests into calls to [`AxisDB`](axisdb/api.py:1).
"""

from __future__ import annotations

from pathlib import Path
from typing import Any

from fastapi import Body, FastAPI, HTTPException, Query

from axisdb import AxisDB
from axisdb.errors import (
    LockError,
    ReadOnlyError,
    StorageCorruptionError,
    ValidationError,
)
from axisdb.query.ast import Field
from axisdb.server.schemas import DeleteBody, InitResponse, ItemBody

app = FastAPI(title="AxisDB")


def _to_http(exc: Exception) -> HTTPException:
    if isinstance(exc, (ValidationError, ReadOnlyError)):
        return HTTPException(status_code=400, detail=str(exc))
    if isinstance(exc, KeyError):
        return HTTPException(status_code=404, detail="Not found")
    if isinstance(exc, LockError):
        return HTTPException(status_code=423, detail=str(exc))
    if isinstance(exc, StorageCorruptionError):
        return HTTPException(status_code=500, detail=str(exc))
    return HTTPException(status_code=500, detail=str(exc))


@app.get("/info")
def info(path: str) -> dict[str, Any]:
    try:
        db = AxisDB.open(path, mode="r")
        return {"path": str(Path(path)), "dimensions": db.dimensions, "mode": "r"}
    except Exception as exc:  # noqa: BLE001
        raise _to_http(exc) from exc


@app.post("/init")
def init_db(path: str, dimensions: int, overwrite: bool = False) -> InitResponse:
    try:
        db = AxisDB.create(path, dimensions=dimensions, overwrite=overwrite)
        return InitResponse(
            path=str(Path(path)),
            dimensions=db.dimensions,
            created=True,
        )
    except Exception as exc:  # noqa: BLE001
        raise _to_http(exc) from exc


@app.post("/item")
def set_item(path: str, body: ItemBody = Body(...)) -> dict[str, Any]:
    try:
        with AxisDB.open(path, mode="rw") as db:
            db.set(tuple(body.coords), body.value)
            db.commit()
        return {"coords": body.coords, "value": body.value}
    except Exception as exc:  # noqa: BLE001
        raise _to_http(exc) from exc


@app.get("/item")
def get_item(path: str, coords: list[str] = Query(...)) -> dict[str, Any]:
    try:
        db = AxisDB.open(path, mode="r")
        value = db.get(tuple(coords))
        return {"coords": coords, "value": value}
    except Exception as exc:  # noqa: BLE001
        raise _to_http(exc) from exc


@app.delete("/item")
def delete_item(path: str, body: DeleteBody = Body(...)) -> dict[str, Any]:
    try:
        with AxisDB.open(path, mode="rw") as db:
            db.delete(tuple(body.coords))
            db.commit()
        return {"coords": body.coords, "deleted": True}
    except Exception as exc:  # noqa: BLE001
        raise _to_http(exc) from exc


@app.get("/list")
def list_items(
    path: str, prefix: list[str] | None = None, depth: int | None = None
) -> dict[str, Any]:
    try:
        db = AxisDB.open(path, mode="r")
        keys = db.list(prefix=tuple(prefix or ()), depth=depth)
        return {"keys": [list(k) for k in keys]}
    except Exception as exc:  # noqa: BLE001
        raise _to_http(exc) from exc


@app.get("/find")
def find_items(
    path: str,
    prefix: list[str] | None = None,
    field: list[str] | None = None,
    op: str = "==",
    value: Any = None,
    limit: int | None = None,
) -> dict[str, Any]:
    """Minimal query endpoint.

    MVP: supports a single field predicate.
    """

    try:
        db = AxisDB.open(path, mode="r")
        expr = None
        if field is not None:
            expr = Field(tuple(field), op, value)  # type: ignore[arg-type]
        rows = db.find(prefix=tuple(prefix or ()), where=expr, limit=limit)
        return {"rows": [{"key": list(k), "value": v} for k, v in rows]}
    except Exception as exc:  # noqa: BLE001
        raise _to_http(exc) from exc
